/* pfdp2.c */

/* This program implements a parallel version of the forward dynamic 
   programming algorithm presented in "Power Generation, Operation, and 
   Control" by A. J. Wood and B. F. Wollenberg.  This program follows 
   the  flowchart in figure 5.5, except that the initialization has been 
   worked into the first time through the loop.  This program limits 
   values of x and n to be equal, and requires no sychronization and 
   sequential calculation between each stage.  This particular program 
   is a specialized version to explicitly work with the diagram in 
   FIG 5.17 on page 140 and problem 5.1 on page 147.  This program is 
   the second proof-of-concept of the parallelized forward dynamic 
   programming algorithm for unit commitment.  The forward dynamic 
   programming algorithm is correct; however, data will be explicitly 
   entered from files for each state in each period rather than 
   calculating the feasible states from unit priority data and load 
   patterns. */

#include <stdio.h>
#include <string.h>
#include <cthreads.h>

/* define some constants */
#define MAXPERIODS    5
#define MAXNODES      4
#define MAXNAMELENGTH 5
#define FILENAME_MAX  24
#define YES           1
#define NO            0
#define MAYBE         1
#define INFINITY      999

/* define a structure that represents a state or node.  a node is
   completely defined by eight items: 
      1. name (name)
      2. least total cost to arrive at this node (fcost)
      3. production cost for this node (pcost)
      4. transition costs from the nodes in the previous period
         to this node (scost[]) 
      5. pointer to the state in the previous period that is
         potentially in the optimal schedule (backpoint)  */

struct node {char             name[MAXNAMELENGTH];
             int              fcost;
             int              pcost;
             int              scost[MAXNODES];
             int              backpoint;
             int              calc};

/* define a matrix of nodes for various periods.  The 0th period is the
   initial state, while there are (m) periods to be examined.  Consequently,
   there are a total of (m+1) periods.  At most (x) nodes will be 
   examined each period. */

struct node nodes[MAXNODES][MAXPERIODS+1];

struct lock_obj {struct mutex     MUTEX;
                 struct condition CONDITION} lock[MAXNODES];

int x;  /* number of states to be examined per period */
int m;  /* number of periods to be examined */

/* define user-supplied functions */
/* function names are designed to be self-explainatory */

void init_mach_local();
void init_constants();
void init_nodes();
void print_nodenames();
void print_scost();
void calc_fcost();
int  fcost();
void signal_fcost();
void trace_optimal_schedule();


/* This program is a parallelized version of the program FDP.C, or the 
   forward dynamic programming example. Operations in the main routine 
   are abstracted by extensive use of functions. */

/* the main program is called with no arguments */

main()
{
 int k;   /* variable that represents the period under examination */
 int q;   /* an index variable */
 int processor_num[MAXNODES];
 int *pq;
 struct t_id {int *thread_id} THREAD[MAXNODES]; /* an array of thread ids 
                                                   to be used when joining 
                                                   the threads */

 init_mach_local();                /* initialize all mach routine */
 init_constants();           /* get constants from the user */
 init_nodes();               /* initialize all nodes from a file */

/* fork the treads for processors 1 - x */

 for (q = 1;q < x;q++)
 {
printf("fork process |%d| \n",q);
fflush(stdout);
  processor_num[q] = q;
  pq = processor_num + q;
/* fork the threads and store the thread id */
  THREAD[q].thread_id = cthread_fork(calc_fcost, pq);
 }

/* calculate fcost on processor 0 */

  processor_num[0] = 0;
  calc_fcost(processor_num);

for (q = 1;q < x;q++)
 {
printf("join process |%d| \n",q);
fflush(stdout);
  cthread_join(THREAD[q].thread_id);   /* join all the threads */
 }
 trace_optimal_schedule(m);     /* trace the optimal schedule */
}


/* This function initializes all mach specific variables */

void init_mach_local()
{
 int i,j,k;  /* index variables */

 cthread_init();  /* initialize cthreads functions */

/* allocate all mutexs and conditionals */
 for (i = 0;i < x;i++)
 {
    mutex_init(&lock[i].MUTEX);
    condition_init(&lock[i].CONDITION);
 }
}


/* This function queries the user for input for constants in the program */

void init_constants()
 {
  printf("input the number of periods to be examined (m) ");
  scanf ("%d",&m);
/* the number of periods cannot be greater than MAXPERIODS */
  if (m > MAXPERIODS)
   {
    m = MAXPERIODS;
    printf("the number of periods cannot be greater than %d \n",MAXPERIODS);
   }

  printf("input the number of states to be examined (x) ");
  scanf ("%d",&x);
/* the number of states cannot be greater than MAXNODES */
  if (x > MAXNODES)
   {
    x = MAXNODES;
    printf("the number of states cannot be greater than %d \n",MAXNODES);
   }
 }


/* This function reads a file to enter the node names and transition
   cost data.  The data is stored with identifier fields for the period
   and the number of the node in the period.  The matrix of nodes for
   the various periods need not be complete.  The only requirements are
   that the user correctly generate the transition costs, and that there
   is at least one node per period.  This input routine was specifically
   designed to handle data similar to FIG. 5.17 or Problem 5.1 in "Power
   Generation, Operation, and Control".

   The routine intializes the matrix of structures before data is entered,
   so that non-assigned states have '*' for a name and the cost to arrive
   at a state and all transition costs are set to INFINITY, a sufficiently
   large integer.  Also for this particular input data, production costs
   are assumed to be zero, so all production values in the matrix of
   structures are assigned to 0.

   The input format, with data ranges, is

           period | node# | node-name | array of scost

   where

   period is an integer and takes values [0-MAXPERIODS]. 
   node# is an integer and takes values [0-(MAXNODES-1)].
   node-name is a character with length less than MAXNAMELENGTH.
   scost values are in an integer array with MAXNODES number of values.  If
   no link exists with a previous node, that value is sent to -1, and the
   program will assign that transition cost to INFINITY.  My appoligies
   for hardwiring the array of scost values into the program, but this is 
   a proof-of-concept program and later programs will use a different 
   scheme to set up node information, when the power generation unit 
   commitment problem is addressed. */

void init_nodes()
 {
  int  i,j,k,q;                /* index variables */
  char filename[FILENAME_MAX]; /* input filename */
  FILE *fp;                    /* pointer to the file stream */
  int  period;                 /* input variable for the period */
  int  nodenum;                /* input variable for the node number */
  char name[MAXNAMELENGTH];    /* input variable for the node name */
/* these scost input variables assume MAXNODES = 4 */
  int  scost0;                 /* input variable for 1st scost */
  int  scost1;                 /* input variable for 2nd scost */
  int  scost2;                 /* input variable for 3rd scost */
  int  scost3;                 /* input variable for 4th scost */

/* initialize the values at each node in the matrix of nodes */
  for (i=0;i<MAXNODES;i++)
   {
    for (j=0;j<=MAXPERIODS;j++)
     {
      nodes[i][j].name[0] = '*';         /* set the node name to a dash */
      nodes[i][j].name[1] = '\0';        /* set the end-of-string marker */
      nodes[i][j].fcost   = INFINITY;    /* set the total cost to INFINITY */
      nodes[i][j].pcost   = 0;           /* set the production cost to 0 */
      for (k=0;k<MAXNODES;k++)
       {
        nodes[i][j].scost[k] = INFINITY; /* set the transition costs from */
       }                                 /* nodes in the previous period */
                                         /* to INFINITY */
      nodes[i][j].backpoint = -1;        /* set the backpointer to an */
                                         /* impossible value */
      nodes[i][j].calc = 0;
     }
   }

/* initialize the costs in the 0th state to 0 */
  for (i=0;i<MAXNODES;i++)
   {
    nodes[i][0].fcost = 0;
    nodes[i][0].calc  = 1;
   }

/* open the node data file */
  printf("input the filename for the node data ");
  scanf ("%s",filename);
  printf("\n");

  fp = fopen(filename, "r"); 
/* read the node data file until an EOF is encountered */

  while ((fscanf(fp,"%d %d %s %d %d %d %d", &period,
                                            &nodenum,
                                            name,
                                            &scost0,
                                            &scost1,
                                            &scost2,
                                            &scost3)) != EOF)
/* copy the input data to the data structure */
   {
    strncpy(nodes[nodenum][period].name,name,3);
    if (scost0 != -1) nodes[nodenum][period].scost[0] = scost0;
    if (scost1 != -1) nodes[nodenum][period].scost[1] = scost1;
    if (scost2 != -1) nodes[nodenum][period].scost[2] = scost2;
    if (scost3 != -1) nodes[nodenum][period].scost[3] = scost3;
   }

  fclose(fp);

/* echo the data to the screen */
  print_nodenames();
  print_scost();
 }


/* This function prints the name and least total cost to arrive at the
   node. */

void print_nodenames()
 {
  int  i,j;  /* index variables */

  printf("\n");
  printf("NODENAME/COST MATRIX \n");

  for (i=0;i<MAXNODES;i++)
   {
    for (j=0;j<=MAXPERIODS;j++)
     {
/* name - cost are represented as a tuple */
      printf("(%2s - %4d)  ",nodes[i][j].name,
                          nodes[i][j].fcost);
     }
    printf("\n");
   }
 }


/* This function prints the transition costs to arrive at a node. */

void print_scost()
 {
  int  i,j,k;  /* index variables */

  printf("\n");
  printf("TRANSITION COST MATRIX \n");

  for (i=0;i<MAXNODES;i++)
   {
    for (k=0;k<MAXNODES;k++)
     {
      for (j=0;j<=MAXPERIODS;j++)
       {
        printf("%4d",nodes[i][j].scost[k]);
       }
      printf("\n");
     }
    printf("\n");
   }
 }


/* This function calculates the minimum cost to arrive at each node
   in a particular period. */

void calc_fcost(pq)
  int *pq;  /* thread identifier*/
 {
  int q = *pq;
  int j,k;    /* index variables */
  int temp;   /* temporary storage value */

  for (k = 1;k <= m;k++)
   {

/* calculate the minimum value of fcost */

    nodes[q][k].fcost = fcost(q,0,k);
    nodes[q][k].backpoint = 0;
    for (j=1;j<x;j++)
     {
      temp = fcost(q,j,k);
      if (temp < nodes[q][k].fcost)
       {
        nodes[q][k].fcost = temp;
        nodes[q][k].backpoint = j;
       }
     }
    signal_fcost(q,k);  /* signal that processing has completed */
   }
 }


/* This function calculates the value of fcost(q,j,k) and waits for 
   data if required.  This fuction returns an integer type value. */

int fcost(q,j,k)
  int q;  /* thread id */
  int j;  /* node id within a stage */
  int k;  /* stage id */
 {
  mutex_lock(&lock[j].MUTEX);  /* lock the mutex */
/* check to see if the node has been calculated */
  while(nodes[j][k-1].calc == 0)
   {
/* wait for the condition to be signaled */
   condition_wait(&lock[j].CONDITION,&lock[j].MUTEX);
   }
  mutex_unlock(&lock[j].MUTEX);  /* unlock the mutex */

/* calculate and return fcost(q,j,k) */
   return (nodes[q][k].pcost +
           nodes[q][k].scost[j] +
           nodes[j][k-1].fcost);
 }

/* This function signals that the value nodes[q][k].fcost has been 
   calculated */

void signal_fcost(q,k)
  int q;  /* thread id */
  int k;  /* stage id */
 {
  int i;  /* index variable */

  nodes[q][k].calc = 1;  /* indicate that the value has been calculated */

    mutex_lock(&lock[q].MUTEX);           /* lock the mutex */
    condition_broadcast(&lock[q].CONDITION); /* signal the condition */
    mutex_unlock(&lock[q].MUTEX);         /* unlock the mutex */
 }


/* This function traces the optimal schedule by starting at the least
   cost final node, and works backward until the 0th state is reached, or
   until a situation is reached where no optimal schedule can be traced.
   This situation could occur when a critical node is discarded.  If
   such a situation would occur, the problem would require reprocessing
   with a larger value of n, or the program could be augmented to include
   a roll-back option that limits the amount of recalculations. */

void trace_optimal_schedule(m)
  int m; /* number of periods to be examined */
 {
  int i,j,k,p;     /* index variables */
  int temp;        /* value used to find the minimum cost final state */
  int test;        /* indicator variables */
  int isave;       /* variable to save the period where the trace failed */
  int os[MAXPERIODS+1]; /* array of nodes in the optimal schedule */

/* echo the data to the screen */
  print_nodenames();

/* find the minimum cost node in the last period */
  p = 0;
  temp = nodes[0][m].fcost;
  for (j=1;j<MAXNODES;j++)
   {
    if (nodes[j][m].fcost < temp)
     {
      p = j;
      temp = nodes[j][m].fcost;
     }
   }
/* record the location of the minimum cost node in the last period */
  os[m] = p;

/* trace through the remaining periods */
/* if test is set to NO, then an optimal schedule cannot be found */
  test = MAYBE;
  for (i=m-1;(i>=0 && test==MAYBE);i--)
   {
    os[i] = nodes[ os[i+1] ][i+1].backpoint;
/* check to see if this was the value used in calculating the optimal
   schedule */
    temp = (nodes[ os[i] ][i].fcost +
            nodes[ os[i+1] ][i+1].pcost +
            nodes[ os[i+1] ][i+1].scost[ os[i] ]);
    if ( os[i] == -1 ||
         nodes[ os[i+1] ][i+1].fcost != temp)
     {
      test = NO;
/* save the period where the trace failed */
      isave = i;
     }
   }

  printf("\n");

/* print information on the trace */
  if (test == YES)
/* an optimal schedule was found */
   {
    printf("OPTIMAL SCHEDULE\n\n");

    for (i=0;i<=m;i++)
     {
      printf("period #%d  (%s - %4d)\n",i,nodes[ os[i] ][i].name,
                                     nodes[ os[i] ][i].fcost);
     }
   }
  else
/* an optimal schedule was not found */
   {
    printf("TRACE WAS UNABLE TO FIND AN OPTIMAL SCHEDULE\n");
    printf("TRACE FAILED IN PERIOD - %d");

    printf("PARTIAL OPTIMAL SCHEDULE\n");
    for (i=isave+1;i<=m;i++)
     {
      printf("period #%d  (%s - %4d)\n",i,nodes[ os[i] ][i].name,
                                     nodes[ os[i] ][i].fcost);
     }
   }
 }

